#!/usr/bin/perl

use strict;
use warnings;

sub assert {
  unless($_[0]) {
    require Carp;
    Carp::confess();
  }
  return undef;
}

sub _OUT {
    my $fd = shift;
    print $fd @_;
}

sub parse_token_code_gen {
    my $fd = shift;

    my $prevpos = shift;	# 0, 1 ...
    my $prev	= shift;	# "G", "GE", ...
    my $term	= shift;	# function(macro) name to check if a symbol is terminator

    assert ((scalar(@_) % 2) == 0);

    my %symbs	= @_;		# symbol => result, ...

    my $pos = $prevpos + 1;

    # label
    if ($prevpos != -1) {
	_OUT ($fd, '	 _'.$prevpos.$prev.":\n");
    }

    my $nsymbs = scalar(keys %symbs);
    assert ($nsymbs);

    if ($nsymbs == 1) {
	my ($m, $r) = %symbs;

	my @cond;
	# optimize for just this one string
	my $i;
	for ($i = $pos; $i < length($m); $i++) {
	    push @cond, '(m['.$i."] == '".substr($m, $i, 1)."')";
	}
	push @cond, '('.$term.'(m['.$i."]))";

	_OUT ($fd, "\t//".$m."\n");
	_OUT ($fd, "\t".'if ('.join(" && ", @cond).') {'."\n");
	_OUT ($fd, "\t".'    r = '.$r.";\n");
	_OUT ($fd, "\t".'    p = '.$i.";\n");
	_OUT ($fd, "\t".'    goto ok;'."\n");
	_OUT ($fd, "\t"."}\n");
	_OUT ($fd, "\t".'goto invalid;'."\n");
    } else {
	my @case;
	my @tocall;

	{
	    ## terminating keys
	    foreach my $m (grep { (length($_) == $pos) } sort keys %symbs) {
		my $r = delete $symbs{$m};

		_OUT ($fd, "\t//".$m."\n");
		_OUT ($fd, "\t".'if ('.$term.'(m['.$pos.'])) {'."\n");
		_OUT ($fd, "\t".'    r = '.$r.";\n");
		_OUT ($fd, "\t".'    p = '.$pos.";\n");
		_OUT ($fd, "\t".'    goto ok;'."\n");
		_OUT ($fd, "\t"."}\n");
	    }
	}
	{
	    ## non-terminating keys
	    my @keys = (sort keys %symbs);
	    # hash by this char
	    my %h;
	    foreach my $k (@keys) {
		assert( length($k) > $pos );
		push @{$h{substr($k, $pos, 1)}}, ($k);
	    }

	    foreach my $k (sort keys %h) {
		my $la = '_'.$pos.$prev.$k;
		my @me = @{$h{$k}};
		push @case, ("case '".$k."':\tgoto ".$la.";\t// ".
			     join(", ",@me));

		push @tocall, [$prev.$k, \@me];
	    }
	}

	push @case, ("default:\tgoto invalid;");

	_OUT ($fd, "\t".'switch (m['.$pos."]) {\n\t");
	_OUT ($fd, join("\n\t", @case)."\n");
	_OUT ($fd, "\t}\n");

	if (scalar(@tocall)) {
	    map {
		parse_token_code_gen($fd, $pos, $_->[0],
				     $term,
				     (map { $_ => $symbs{$_}; }
				      @{$_->[1]}));
	    } @tocall;
	}
    }
}

my @vals = (qw(
	IDENTITY
	BASE64
	BASE64URL
	BASE64URLNOPAD
	HEX
	HEXUC
	HEXLC));

unless ($#ARGV == 1) {
    printf(STDERR 'Usage: '.$0." <c-file> <h-file>\n");
    exit 1;
}

my ($c_file, $h_file) = @ARGV;

open(my $c_out, '>', $c_file) || die('cannot open '.$c_file.': '.$!);
open(my $h_out, '>', $h_file) || die('cannot open '.$h_file.': '.$!);


_OUT($h_out, <<'EOF');

/* AUTO-GENERATED, DO NOT EDIT */

EOF

_OUT($h_out,
     "enum encoding {\n\t_INVALID = 0,\n\t".
     join(",\n\t", @vals).",\n\t__MAX_ENCODING\n};\n");

_OUT($h_out, <<'EOF');

enum encoding parse_encoding (const char *);

EOF

_OUT($c_out, <<EOF);

/* AUTO-GENERATED, DO NOT EDIT */

#include "$h_file"
#define term(c) ((c) == '\\0')

EOF

_OUT($c_out, <<'EOF');
enum encoding
parse_encoding (const char *m) {
	int p;
	enum encoding r;

EOF

parse_token_code_gen($c_out, -1, '', 'term',
		     map { ($_ => $_) } @vals);

_OUT($c_out, <<EOF);
  ok:
	return r;
  invalid:
    return _INVALID;
    (void)p;
}
EOF

close($c_out) || die('error closing '.$c_file.': '.$!);
close($h_out) || die('error closing '.$h_file.': '.$!);
